"""Tests for hussh.connection module."""

from pathlib import Path

import pytest

from hussh import Connection, SSHResult

TEXT_FILE = Path("tests/data/hp.txt").resolve()
IMG_FILE = Path("tests/data/puppy.jpeg").resolve()


@pytest.fixture
def conn():
    """Return a basic Connection object."""
    return Connection(host="localhost", port=8022, password="toor")


def test_password_auth():
    """Test that we can establish a connection with password-based authentication."""
    assert Connection(host="localhost", port=8022, password="toor")


def test_key_auth():
    """Test that we can establish a connection with key-based authentication."""
    assert Connection(host="localhost", port=8022, private_key="tests/data/test_key")


def test_key_with_password_auth():
    """Test that we can establish a connection with key-based authentication and a password."""
    assert Connection(
        host="localhost",
        port=8022,
        private_key="tests/data/auth_test_key",
        password="husshpuppy",
    )


@pytest.mark.skip("fixture-based setup for agent-based auth currently not working")
def test_agent_auth(setup_agent_auth):
    """Test that we can establish a connection with agent-based authentication."""
    assert Connection(host="localhost", port=8022)


def test_basic_command(conn):
    """Test that we can run a basic command."""
    result = conn.execute("echo hello")
    assert isinstance(result, SSHResult)
    assert result.status == 0
    assert result.stdout == "hello\n"


def test_bad_command(conn):
    """Test that we can run a bad command."""
    result = conn.execute("kira")
    assert result.status != 0
    assert "command not found" in result.stderr


def test_text_scp(conn):
    """Test that we can copy a file to the server and read it back."""
    # copy a local file to the server
    conn.scp_write(str(TEXT_FILE), "/root/hp.txt")
    assert "hp.txt" in conn.execute("ls /root").stdout
    # read the file back from the server
    read_text = conn.scp_read("/root/hp.txt")
    hp_text = Path(str(TEXT_FILE)).read_text()
    assert read_text == hp_text
    # copy the file from the server to a local file
    conn.scp_read("/root/hp.txt", "scp_hp.txt")
    scp_hp_text = Path("scp_hp.txt").read_text()
    Path("scp_hp.txt").unlink()
    assert scp_hp_text == hp_text


def test_scp_write_data(conn):
    """Test that we can write a string to a file on the server."""
    conn.scp_write_data("hello", "/root/hello.txt")
    assert "hello.txt" in conn.execute("ls /root").stdout
    read_text = conn.scp_read("/root/hello.txt")
    assert read_text == "hello"


@pytest.mark.skip("non-text files are not supported by scp")
def test_non_utf8_scp(conn):
    """Test that we can copy a non-text file to the server and read it back."""
    # copy an image file to the server
    conn.scp_write(str(IMG_FILE), "/root/puppy.jpeg")
    assert "puppy.jpeg" in conn.execute("ls /root").stdout
    # read the file back from the server
    read_img = conn.scp_read("/root/puppy.jpeg")
    img_data = Path(str(IMG_FILE)).read_bytes()
    assert read_img == img_data
    # copy the file from the server to a local file
    conn.scp_read("/root/puppy.jpeg", "scp_puppy.jpeg")
    scp_img_data = Path("scp_puppy.jpeg").read_bytes()
    Path("scp_puppy.jpeg").unlink()
    assert scp_img_data == img_data


def test_text_sftp(conn):
    """Test that we can copy a file to the server and read it back."""
    # copy a local file to the server
    conn.sftp_write(str(TEXT_FILE), "/root/hp.txt")
    assert "hp.txt" in conn.execute("ls /root").stdout
    # read the file back from the server
    read_text = conn.sftp_read("/root/hp.txt")
    hp_text = Path(str(TEXT_FILE)).read_text()
    assert read_text == hp_text
    # copy the file from the server to a local file
    conn.sftp_read("/root/hp.txt", "sftp_hp.txt")
    sftp_hp_text = Path("sftp_hp.txt").read_text()
    Path("sftp_hp.txt").unlink()
    assert sftp_hp_text == hp_text


def test_sftp_write_data(conn):
    """Test that we can write a string to a file on the server."""
    conn.sftp_write_data("hello", "/root/hello.txt")
    assert "hello.txt" in conn.execute("ls /root").stdout
    read_text = conn.sftp_read("/root/hello.txt")
    assert read_text == "hello"


@pytest.mark.skip("non-text files are not supported by sftp")
def test_non_utf8_sftp(conn):
    """Test that we can copy a non-text file to the server and read it back."""
    # copy an image file to the server
    conn.sftp_write(str(IMG_FILE), "/root/puppy.jpeg")
    assert "puppy.jpeg" in conn.execute("ls /root").stdout
    # read the file back from the server
    read_img = conn.sftp_read("/root/puppy.jpeg")
    img_data = Path(str(IMG_FILE)).read_bytes()
    assert read_img == img_data
    # copy the file from the server to a local file
    conn.sftp_read("/root/puppy.jpeg", "sftp_puppy.jpeg")
    sftp_img_data = Path("sftp_puppy.jpeg").read_bytes()
    Path("sftp_puppy.jpeg").unlink()
    assert sftp_img_data == img_data


def test_shell_context(conn):
    """Test that we can run multiple commands in a shell context."""
    with conn.shell() as sh:
        sh.send("echo test shell")
        sh.send("bad command")
    assert "test shell" in sh.result.stdout
    assert "command not found" in sh.result.stderr
    assert sh.result.status != 0


def test_pty_shell_context(conn):
    """Test that we can run multiple commands in a pty shell context."""
    with conn.shell(pty=True) as sh:
        sh.send("echo test shell")
        sh.send("bad command")
    assert "test shell" in sh.result.stdout
    assert "command not found" in sh.result.stdout
    assert sh.result.status != 0


def test_session_timeout():
    """Test that we can trigger a timeout on session handshake."""
    with pytest.raises(TimeoutError):
        Connection(host="localhost", port=8022, password="toor", timeout=10)


def test_command_timeout(conn):
    """Test that we can trigger a timeout on command execution."""
    with pytest.raises(TimeoutError):
        conn.execute("sleep 5", timeout=3000)


def test_remote_copy(conn, run_second_server):
    """Test that we can copy a file from one server to another."""
    # First copy the test file to the first server
    conn.scp_write(str(TEXT_FILE), "/root/hp.txt")
    assert "hp.txt" in conn.execute("ls /root").stdout
    # Now copy the file from the first server to the second server
    dest_conn = Connection(host="localhost", port=8023, password="toor")
    conn.remote_copy("/root/hp.txt", dest_conn)
    assert "hp.txt" in dest_conn.execute("ls /root").stdout


def test_tail(conn):
    """Test that we can tail a file."""
    TEST_STR = "hello\nworld\n"
    conn.scp_write_data(TEST_STR, "/root/hello.txt")
    with conn.tail("/root/hello.txt") as tf:
        assert tf.read(0) == TEST_STR
        assert tf.last_pos == len(TEST_STR)
        conn.execute("echo goodbye >> /root/hello.txt")
    assert tf.contents == "goodbye\n"



import json
from pathlib import Path
from pprint import pprint
import timeit

import memray

results_dict = {}

if (mem_path := Path("memray-bench_fabric.bin")).exists():
    mem_path.unlink()
with memray.Tracker("memray-bench_fabric.bin", native_traces=True, follow_fork=True):
    start_time = timeit.default_timer()

    from fabric import Connection

    results_dict["import_time"] = f"{(timeit.default_timer() - start_time) * 1000:.2f} ms"
    host_info = json.loads(Path("target.json").read_text())

    temp_time = timeit.default_timer()
    conn = Connection(
        host=host_info["host"],
        port=host_info["port"],
        user=host_info["username"],
        connect_kwargs={
            "password": host_info["password"],
            "look_for_keys": False,
            "allow_agent": False,
        },
    )
    conn.open()
    results_dict["connect_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"

    temp_time = timeit.default_timer()
    result = conn.run("echo test")
    results_dict["cmd_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"

    # small file (1kb)
    temp_time = timeit.default_timer()
    conn.put("1kb.txt", "/root/1kb.txt")
    results_dict["s_put_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"

    temp_time = timeit.default_timer()
    conn.get("/root/1kb.txt", "small.txt")
    results_dict["s_get_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"
    Path("small.txt").unlink()

    # medium file (14kb)
    temp_time = timeit.default_timer()
    conn.put("14kb.txt", "/root/14kb.txt")
    results_dict["m_put_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"

    temp_time = timeit.default_timer()
    conn.get("/root/14kb.txt", "medium.txt")
    results_dict["m_get_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"
    Path("medium.txt").unlink()

    # large file (64kb)
    temp_time = timeit.default_timer()
    conn.put("64kb.txt", "/root/64kb.txt")
    results_dict["l_put_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"

    temp_time = timeit.default_timer()
    conn.get("/root/64kb.txt", "large.txt")
    results_dict["l_get_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"
    Path("large.txt").unlink()

    conn.close()

    results_dict["total_time"] = f"{(timeit.default_timer() - start_time) * 1000:.2f} ms"

pprint(results_dict, sort_dicts=False)

if Path("bench_results.json").exists():
    results = json.loads(Path("bench_results.json").read_text())
else:
    results = {}
results.update({"fabric": results_dict})
Path("bench_results.json").write_text(json.dumps(results, indent=2))



"""Common setup for Hussh tests."""

import os
from pathlib import PurePath
import subprocess
import time

import docker
import pexpect
import pytest

TESTDIR = PurePath(__file__).parent


@pytest.fixture(scope="session")
def ensure_test_server_image():
    """Ensure that the test server Docker image is available."""
    client = docker.from_env()
    try:
        client.images.get("hussh-test-server")
    except docker.errors.ImageNotFound:
        client.images.build(
            path=str(TESTDIR / "setup"),
            tag="hussh-test-server",
        )
    client.close()


@pytest.fixture(scope="session", autouse=True)
def run_test_server(ensure_test_server_image):
    """Run a test server in a Docker container."""
    client = docker.from_env()
    try:  # check to see if the container is already running
        container = client.containers.get("hussh-test-server")
    except docker.errors.NotFound:  # if not, start it
        container = client.containers.run(
            "hussh-test-server",
            detach=True,
            ports={"22/tcp": 8022},
            name="hussh-test-server",
        )
        time.sleep(5)  # give the server time to start
    yield container
    container.stop()
    container.remove()
    client.close()


@pytest.fixture(scope="session")
def run_second_server(ensure_test_server_image):
    """Run a test server in a Docker container."""
    client = docker.from_env()
    try:  # check to see if the container is already running
        container = client.containers.get("hussh-test-server2")
    except docker.errors.NotFound:  # if not, start it
        container = client.containers.run(
            "hussh-test-server",
            detach=True,
            ports={"22/tcp": 8023},
            name="hussh-test-server2",
        )
        time.sleep(5)  # give the server time to start
    yield container
    container.stop()
    container.remove()
    client.close()


@pytest.fixture(scope="session")
def setup_agent_auth():
    # Define the key paths
    base_key = TESTDIR / "data/test_key"
    auth_key = TESTDIR / "data/auth_test_key"

    # Start the ssh-agent and get the environment variables
    output = subprocess.check_output(["ssh-agent", "-s"])
    env = dict(line.split("=", 1) for line in output.decode().splitlines() if "=" in line)

    # Set the SSH_AUTH_SOCK and SSH_AGENT_PID environment variables
    os.environ["SSH_AUTH_SOCK"] = env["SSH_AUTH_SOCK"]
    os.environ["SSH_AGENT_PID"] = env["SSH_AGENT_PID"]

    # Add the keys to the ssh-agent
    # subprocess.run(["ssh-add", str(base_key)], check=True)
    result = subprocess.run(
        ["ssh-add", str(base_key)], capture_output=True, text=True, check=False
    )
    print(result.stdout)
    print(result.stderr)
    # The auth_key is password protected
    child = pexpect.spawn("ssh-add", [str(auth_key)])
    child.expect("Enter passphrase for .*: ")
    child.sendline("husshpuppy")
    yield
    # Kill the ssh-agent after the tests have run
    subprocess.run(["ssh-agent", "-k"], check=True)



import json
from pathlib import Path
import subprocess

from rich.console import Console
from rich.table import Table


def run_all():
    """Find all the python files in this directory starting with bench_ and run them."""
    for file in Path(__file__).parent.glob("bench_*.py"):
        print(f"Running {file}")
        subprocess.run(["python", file], check=True)


def run_memray_reports(report_dict):
    """Find all memray reports, run them, then delete them."""
    for file in Path(__file__).parent.glob("memray-*.bin"):
        # Figure out what library we're looking at
        lib = file.stem.replace("memray-bench_", "")
        json_file = Path(f"{lib}.json")
        subprocess.run(["memray", "stats", "--json", file, "-o", str(json_file)], check=True)
        file.unlink()
        # load the new json file
        results = json.loads(json_file.read_text())
        report_dict[lib]["peak_memory"] = (
            f'{results["metadata"]["peak_memory"] / 1024 / 1024:.2f} MB'
        )
        report_dict[lib]["allocations"] = str(results["metadata"]["total_allocations"])
        json_file.unlink()


def print_report(report_dict):
    """Print out the report in a rich table"""
    report_table = Table(title="Benchmark Report")
    # Add the columns
    report_table.add_column("Library")
    for key in report_dict["hussh"]:
        report_table.add_column(key)
    for lib in report_dict:
        report_table.add_row(lib, *[report_dict[lib][key] for key in report_dict[lib]])
    Console().print(report_table)


if __name__ == "__main__":
    run_all()
    report_dict = json.loads(Path("bench_results.json").read_text())
    run_memray_reports(report_dict)
    print_report(report_dict)
    Path("bench_results.json").unlink()



import json
from pathlib import Path
from pprint import pprint
import timeit

import memray

results_dict = {}

if (mem_path := Path("memray-bench_ssh2-python.bin")).exists():
    mem_path.unlink()
with memray.Tracker("memray-bench_ssh2-python.bin"):
    start_time = timeit.default_timer()
    import socket

    from ssh2 import sftp
    from ssh2.session import Session

    results_dict["import_time"] = f"{(timeit.default_timer() - start_time) * 1000:.2f} ms"

    host_info = json.loads(Path("target.json").read_text())

    # connect to the server
    temp_time = timeit.default_timer()
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect((host_info["host"], host_info["port"]))
    session = Session()
    session.handshake(sock)
    session.userauth_password(host_info["username"], host_info["password"])
    results_dict["connect_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"

    # execute a command
    temp_time = timeit.default_timer()
    channel = session.open_session()
    channel.execute("echo test")
    channel.wait_eof()
    channel.close()
    channel.wait_closed()
    size, data = channel.read()
    stdout = ""
    while size > 0:
        stdout += data.decode("utf-8")
        size, data = channel.read()
    stderr = channel.read_stderr()
    status = channel.get_exit_status()
    results_dict["cmd_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"

    # small file (1kb)
    temp_time = timeit.default_timer()
    SFTP_MODE = (
        sftp.LIBSSH2_SFTP_S_IRUSR
        | sftp.LIBSSH2_SFTP_S_IWUSR
        | sftp.LIBSSH2_SFTP_S_IRGRP
        | sftp.LIBSSH2_SFTP_S_IROTH
    )
    FILE_FLAGS = sftp.LIBSSH2_FXF_CREAT | sftp.LIBSSH2_FXF_WRITE | sftp.LIBSSH2_FXF_TRUNC
    data = Path("1kb.txt").read_bytes()
    sftp_conn = session.sftp_init()
    with sftp_conn.open("/root/1kb.txt", FILE_FLAGS, SFTP_MODE) as f:
        f.write(data)
    results_dict["s_put_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"

    temp_time = timeit.default_timer()
    with sftp_conn.open("/root/1kb.txt", sftp.LIBSSH2_FXF_READ, sftp.LIBSSH2_SFTP_S_IRUSR) as f:
        read_data = b""
        for _rc, data in f:
            read_data += data
    Path("small.txt").write_bytes(read_data)
    results_dict["s_get_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"
    Path("small.txt").unlink()

    # medium file (14kb)
    temp_time = timeit.default_timer()
    data = Path("14kb.txt").read_bytes()
    with sftp_conn.open("/root/14kb.txt", FILE_FLAGS, SFTP_MODE) as f:
        f.write(data)
    results_dict["m_put_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"

    temp_time = timeit.default_timer()
    with sftp_conn.open("/root/14kb.txt", sftp.LIBSSH2_FXF_READ, sftp.LIBSSH2_SFTP_S_IRUSR) as f:
        read_data = b""
        for _rc, data in f:
            read_data += data
    Path("medium.txt").write_bytes(read_data)
    results_dict["m_get_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"
    Path("medium.txt").unlink()

    # large file (64kb)
    temp_time = timeit.default_timer()
    data = Path("64kb.txt").read_bytes()
    with sftp_conn.open("/root/14kb.txt", FILE_FLAGS, SFTP_MODE) as f:
        f.write(data)
    results_dict["l_put_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"

    temp_time = timeit.default_timer()
    with sftp_conn.open("/root/64kb.txt", sftp.LIBSSH2_FXF_READ, sftp.LIBSSH2_SFTP_S_IRUSR) as f:
        read_data = b""
        for _rc, data in f:
            read_data += data
    Path("large.txt").write_bytes(read_data)
    results_dict["l_get_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"
    Path("large.txt").unlink()

    results_dict["total_time"] = f"{(timeit.default_timer() - start_time) * 1000:.2f} ms"

pprint(results_dict, sort_dicts=False)

if Path("bench_results.json").exists():
    results = json.loads(Path("bench_results.json").read_text())
else:
    results = {}
results.update({"ssh2-python": results_dict})
Path("bench_results.json").write_text(json.dumps(results, indent=2))



import json
from pathlib import Path
from pprint import pprint
import timeit

import memray

results_dict = {}

if (mem_path := Path("memray-bench_hussh.bin")).exists():
    mem_path.unlink()
with memray.Tracker("memray-bench_hussh.bin"):
    start_time = timeit.default_timer()
    from hussh import Connection

    results_dict["import_time"] = f"{(timeit.default_timer() - start_time) * 1000:.2f} ms"

    host_info = json.loads(Path("target.json").read_text())

    temp_time = timeit.default_timer()
    conn = Connection(
        host=host_info["host"],
        port=host_info["port"],
        password=host_info["password"],
    )
    results_dict["connect_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"

    temp_time = timeit.default_timer()
    result = conn.execute("echo test")
    results_dict["cmd_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"

    # small file (1kb)
    temp_time = timeit.default_timer()
    conn.sftp_write("1kb.txt", "/root/1kb.txt")
    results_dict["s_put_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"

    temp_time = timeit.default_timer()
    conn.sftp_read("/root/1kb.txt", "small.txt")
    results_dict["s_get_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"
    Path("small.txt").unlink()

    # medium file (14kb)
    temp_time = timeit.default_timer()
    conn.sftp_write("14kb.txt", "/root/14kb.txt")
    results_dict["m_put_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"

    temp_time = timeit.default_timer()
    conn.sftp_read("/root/14kb.txt", "medium.txt")
    results_dict["m_get_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"
    Path("medium.txt").unlink()

    # large file (64kb)
    temp_time = timeit.default_timer()
    conn.sftp_write("64kb.txt", "/root/64kb.txt")
    results_dict["l_put_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"

    temp_time = timeit.default_timer()
    conn.sftp_read("/root/64kb.txt", "large.txt")
    results_dict["l_get_time"] = f"{(timeit.default_timer() - temp_time) * 1000:.2f} ms"
    Path("large.txt").unlink()

    results_dict["total_time"] = f"{(timeit.default_timer() - start_time) * 1000:.2f} ms"

pprint(results_dict, sort_dicts=False)

if Path("bench_results.json").exists():
    results = json.loads(Path("bench_results.json").read_text())
else:
    results = {}
results.update({"hussh": results_dict})
Path("bench_results.json").write_text(json.dumps(results, indent=2))
